;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "python-coverage" "1.1.0"
  "Show Python coverage via overlays or Flycheck."
  '((emacs "25.1")
    (dash  "2.18.0")
    (s     "1.12.0")
    (xml+  "0"))
  :url "https://github.com/wbolster/emacs-python-coverage"
  :commit "9e7af6999077ed3f861c5813297aee13804fcfb1"
  :revdesc "1.1.0-0-g9e7af6999077"
  :keywords '("languages" "processes" "tools")
  :authors '(("wouter bolsterlee" . "wouter@bolsterl.ee"))
  :maintainers '(("wouter bolsterlee" . "wouter@bolsterl.ee")))
